"""Helper classes and methods to assist with the general function of this bot.

Fedinesia - deletes old statuses for a fediverse account (Mastodon or Pleroma and forks)
Copyright (C) 2021, 2022, 2023  Mark S Burgunder

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import json
import logging
import os
import sys
from dataclasses import asdict
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Optional
from typing import Type
from typing import TypeVar

import aiohttp
from minimal_activitypub import __display_name__ as ma_log_name
from minimal_activitypub.client_2_server import ActivityPub
from minimal_activitypub.client_2_server import ActivityPubError
from rich import print as rprint
from rich.logging import RichHandler

from . import CLIENT_WEBSITE
from . import USER_AGENT
from . import __display_name__

BC = TypeVar("BC", bound="BotConfig")
ConfigClass = TypeVar("ConfigClass", bound="Configuration")
MC = TypeVar("MC", bound="MastodonConfig")

logger = logging.getLogger(__display_name__)


@dataclass
class BotConfig:
    """Dataclass holding configuration values for general behaviour of
    Fedinesia.
    """

    log_level: str
    delete_after: int
    skip_deleting_pinned: Optional[bool]
    skip_deleting_faved: Optional[bool]
    skip_deleting_bookmarked: Optional[bool]
    skip_deleting_poll: Optional[bool]
    skip_deleting_dm: Optional[bool]
    skip_deleting_media: Optional[bool]
    skip_deleting_faved_at_least: Optional[int]
    skip_deleting_boost_at_least: Optional[int]
    skip_deleting_reactions_at_least: Optional[int]

    def __init__(self: BC, config: Optional[Dict[str, Any]]) -> None:  # noqa: C901
        """Initialise instance."""
        if not config:
            config = {}

        self.log_level = config.get("log_level", "WARN")

        std_out_formatter = logging.Formatter(
            "%(name)s[%(process)d] %(levelname)s %(message)s",
            datefmt="%H:%M:%S",
        )
        std_out_handler = RichHandler()
        std_out_handler.setFormatter(std_out_formatter)
        std_out_handler.setLevel(self.log_level)
        logger.addHandler(std_out_handler)

        self.delete_after = config.get("delete_after", None)
        self.skip_deleting_bookmarked = config.get("skip_deleting_bookmarked")
        self.skip_deleting_faved = config.get("skip_deleting_faved")
        self.skip_deleting_pinned = config.get("skip_deleting_pinned")
        self.skip_deleting_poll = config.get("skip_deleting_poll")
        self.skip_deleting_dm = config.get("skip_deleting_dm")
        self.skip_deleting_media = config.get("skip_deleting_media")
        self.skip_deleting_faved_at_least = config.get("skip_deleting_faved_at_least")
        self.skip_deleting_boost_at_least = config.get("skip_deleting_boost_at_least")
        self.skip_deleting_reactions_at_least = config.get(
            "skip_deleting_reactions_at_least"
        )

        if not self.delete_after:
            self._get_delete_after()

        if self.skip_deleting_bookmarked is None:
            self._get_skip_bookmarked()

        if self.skip_deleting_faved is None:
            self._get_skip_faved()

        if self.skip_deleting_pinned is None:
            self._get_skip_pinned()

        if self.skip_deleting_poll is None:
            self._get_skip_poll()

        if self.skip_deleting_dm is None:
            self._get_skip_dm()

        if self.skip_deleting_media is None:
            self._get_skip_media()

        if self.skip_deleting_faved_at_least is None:
            self._get_skip_faved_at_least()

        if self.skip_deleting_boost_at_least is None:
            self._get_skip_boost_at_lease()

        if self.skip_deleting_reactions_at_least is None:
            self._get_skip_reactions_at_lease()

    def _get_skip_poll(self: BC) -> None:
        """Private method to get skip deleting polls value from user if this
        value has not yet been configured.
        """
        rprint("Should polls be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_poll = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_poll = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_dm(self: BC) -> None:
        """Private method to get skip deleting 'private' messages value from
        user if this value has not yet been configured.
        """
        rprint("Should Direct Messages be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_dm = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_dm = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_media(self: BC) -> None:
        """Private method to get skip deleting statuses with media value from
        user if this value has not yet been configured.
        """
        rprint(
            "Should Statuses with attachments / pictures be deleted "
            "when they get old enough?"
        )
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_media = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_media = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_faved_at_least(self: BC) -> None:
        """Private method to get skip deleting statuses that have been
        favourited value from user if this value has not yet been
        configured.
        """
        rprint(
            "Should statuses being favourited a certain minimum number of times be "
            "excluded from deletion even when they get old enough?"
        )
        rprint("(enter 0 to disregard this setting)")
        self.skip_deleting_faved_at_least = int(input("[..] Please enter number: "))

    def _get_skip_boost_at_lease(self: BC) -> None:
        """Private method to get skip deleting statuses that have been
        boosted value from user if this value has not yet been configured.
        """
        rprint(
            "Should statuses being boosted a certain minimum number of times be "
            "excluded from deletion even when they get old enough?"
        )
        rprint("(enter 0 to disregard this setting)")
        self.skip_deleting_boost_at_least = int(input("[..] Please enter number: "))

    def _get_skip_reactions_at_lease(self: BC) -> None:
        """Private method to get skip deleting statuses that have been
        boosted value from user if this value has not yet been configured.
        """
        rprint(
            "Should statuses with a certain minimum number of reactions be "
            "excluded from deletion even when they get old enough?"
        )
        rprint("(enter 0 to disregard this setting)")
        self.skip_deleting_reactions_at_least = int(input("[..] Please enter number: "))

    def _get_skip_pinned(self: BC) -> None:
        """Private method to get skip deleting pinned statuses value from
        user if this value has not yet been configured.
        """
        rprint("Should pinned statuses be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_pinned = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_pinned = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_faved(self: BC) -> None:
        """Private method to get skip deleting favorited statuses value from
        user if this value has not yet been configured.
        """
        rprint("Should favoured statuses be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_faved = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_faved = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_bookmarked(self: BC) -> None:
        """Private method to get skip deleting bookmarked statuses from user
        if this value has not yet been configured.
        """
        rprint("Should bookmarked statuses be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_bookmarked = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_bookmarked = True
        else:
            rprint("! ERROR ... please only respond with 'Y' or 'N'")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_delete_after(self: BC) -> None:
        """Private method to get delete after value from user if this value
        has not yet been configured.
        """
        rprint(
            "Please enter maximum age of retained statuses in the "
            'format of "number unit"'
        )
        rprint('For example "1 weeks" or "3 days". Supported units are:')
        rprint(
            " - seconds\n"
            " - minutes\n"
            " - hours\n"
            " - days\n"
            " - weeks\n"
            " - months"
        )
        max_age = input("[..] Minimum age to delete statuses (in seconds): ")
        max_age_parts = max_age.split(" ")
        max_age_number = int(max_age_parts[0])
        max_age_unit = max_age_parts[1]
        if max_age_unit == "seconds":
            self.delete_after = max_age_number
        elif max_age_unit == "minutes":
            self.delete_after = max_age_number * 3600
        elif max_age_unit == "hours":
            self.delete_after = max_age_number * 3600
        elif max_age_unit == "days":
            self.delete_after = max_age_number * 3600 * 24
        elif max_age_unit == "weeks":
            self.delete_after = max_age_number * 3600 * 24 * 7
        elif max_age_unit == "months":
            self.delete_after = max_age_number * 3600 * 24 * 30
        else:
            rprint("! Error ... unknown unit ({max_age_unit}) specified")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)


@dataclass
class MastodonConfig:
    """Dataclass holding configuration values for Mastodon settings."""

    instance: str
    access_token: str

    @classmethod
    async def load_mastodon_config(
        cls: Type[MC], mastodon_ini: Optional[Dict[str, Any]]
    ) -> MC:
        """Load Mastodon configuration from file.

        :returns:
            MastodonConfig instance
        """
        logger.debug(
            "MastodonConfig.load_mastodon_config - mastodon_ini: %s", mastodon_ini
        )

        if not mastodon_ini:
            mastodon_ini = {}

        base_url = mastodon_ini.get("base_url", None)
        instance = mastodon_ini.get("instance", None)
        if instance is None and base_url:
            instance = base_url

        if not instance:
            instance = input(
                "[..] Enter instance (domain name) for Mastodon account host: "
            )

        access_token = mastodon_ini.get("access_token", None)

        user_info = mastodon_ini.get("user_info", None)
        if not access_token and user_info:
            access_token = user_info.get("access_token", None)

        try:
            if not access_token:
                async with aiohttp.ClientSession() as session:
                    # Create app
                    client_id, client_secret = await ActivityPub.create_app(
                        instance_url=instance,
                        session=session,
                        user_agent=USER_AGENT,
                        client_website=CLIENT_WEBSITE,
                    )

                    # Get Authorization Code / URL
                    authorization_request_url = (
                        await ActivityPub.generate_authorization_url(
                            instance_url=instance,
                            client_id=client_id,
                            user_agent=USER_AGENT,
                        )
                    )
                    print(
                        f"Please go to the following URL and follow the instructions:\n"
                        f"{authorization_request_url}"
                    )
                    authorization_code = input(
                        "[...] Please enter the authorization code:"
                    )

                    # Validate authorization code and get access token
                    access_token = await ActivityPub.validate_authorization_code(
                        session=session,
                        instance_url=instance,
                        authorization_code=authorization_code,
                        client_id=client_id,
                        client_secret=client_secret,
                    )

        except ActivityPubError as error:
            rprint(f"! Error when setting up Fediverse connection: {error}")
            rprint("! Cannot continue. Exiting now.")
            sys.exit(1)

        return cls(instance=instance, access_token=access_token)


@dataclass
class Configuration:
    """Dataclass to hold all settings for fedinesia."""

    bot: BotConfig
    mastodon: MastodonConfig

    @classmethod
    async def load_config(
        cls: Type[ConfigClass], config_file_name: str = "config.json"
    ) -> ConfigClass:
        """Load configuration from file.

        :returns:
            Configuration instance
        """
        if os.path.exists(config_file_name):
            with open(file=config_file_name, encoding="UTF-8") as config_file:
                loaded_config = json.load(config_file)
        else:
            loaded_config = {}

        mastodon_ini = loaded_config.get("mastodon")
        if not mastodon_ini:
            mastodon_ini = loaded_config.get("Mastodon")
        mastodon_config = await MastodonConfig.load_mastodon_config(
            mastodon_ini=mastodon_ini
        )

        bot_ini = loaded_config.get("bot")
        if not bot_ini:
            bot_ini = loaded_config.get("Bot")
        bot = BotConfig(config=bot_ini)

        config_instance = cls(bot=bot, mastodon=mastodon_config)
        config_dict = asdict(config_instance)
        with open(file=config_file_name, mode="w", encoding="UTF-8") as config_file:
            json.dump(config_dict, config_file, indent=4)
            logger.debug("Configuration.load_config - Saved config: %s", config_dict)

        return config_instance


async def setup_shop(
    config_file: str,
    debug_log_file: Optional[str],
) -> Configuration:
    """Process command line arguments, establish debug logging to file if
    specified, load configuration.

    :returns:
        Configuration: config for this run of the Fedinesia.
    """
    file_log_formatter = logging.Formatter(
        "%(asctime)s %(name)s[%(process)d] %(levelname)s %(message)s",
        datefmt="%H:%M:%S",
    )
    if debug_log_file:
        file_handler = logging.FileHandler(filename=debug_log_file)
        file_handler.setFormatter(file_log_formatter)
        file_handler.setLevel(logging.DEBUG)
        logger.addHandler(file_handler)
        ma_logger = logging.getLogger(ma_log_name)
        ma_logger.setLevel(logging.DEBUG)
        ma_logger.addHandler(file_handler)

    config = await Configuration.load_config(config_file_name=config_file)
    return config



"""Fedinesia - deletes old statuses for a fediverse account (Mastodon or Pleroma and forks)
Copyright (C) 2021, 2022, 2023  Mark S Burgunder.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import csv
import json
import logging
import os
from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import cast

import aiocsv
import aiofiles
import arrow
from outdated import check_outdated
from rich import print

from . import Status
from . import __display_name__
from . import __package_name__
from . import __version__
from .config import Configuration

logger = logging.getLogger(__display_name__)


def check_updates() -> None:
    """Check if there is a newer version of Fedinesia available on PyPI."""
    is_outdated = False
    try:
        is_outdated, pypi_version = check_outdated(
            package=__package_name__,
            version=__version__,
        )
        if is_outdated:
            print(
                f"[bold][red]!!! New version of {__display_name__} ({pypi_version}) "
                f"is available on PyPI.org !!!\n"
            )
            logger.debug(
                "check_updates - New version of %s (%s) is available on PyPI.org",
                __display_name__,
                pypi_version,
            )
    except ValueError:
        print(
            "[yellow]Notice - Your version is higher than last published version on PyPI"
        )
        logger.debug(
            "check_updates - Notice - Your version is higher than last published version on PyPI"
        )
    except KeyError:
        logger.debug("check_updates() -> Project not found.")


AL = TypeVar("AL", bound="AuditLog")


@dataclass
class AuditLog:
    """Class to control the creation of an audit log and adding entries to
    it.
    """

    class Style(str, Enum):
        """Enumerating different audit log file styles implemented."""

        PLAIN = "PLAIN"
        CSV = "CSV"

    audit_log: Any
    style: Style

    @classmethod
    async def open(
        cls: Type[AL],
        audit_log_file: str,
        style: Optional[str],
    ) -> AL:
        """Open audit log file for writing and appending new log entries.

        :returns: a new AuditLog instance.
        """
        audit_log = await aiofiles.open(file=audit_log_file, mode="at")
        if style:
            log_style = AuditLog.Style[style]
        else:
            log_style = AuditLog.Style.PLAIN

        if (
            log_style == AuditLog.Style.CSV
            and os.fstat(audit_log.fileno()).st_size == 0
        ):
            await AuditLog._add_csv_header(audit_log)

        return cls(audit_log=audit_log, style=log_style)

    async def close(self: AL) -> None:
        """Close audit log file."""
        self.audit_log.close()

    async def add_entry(self: AL, status: Status) -> None:
        """Append an entry/status details to the audit log file."""
        if self.style == AuditLog.Style.PLAIN:
            await self._add_plain_entry(status)
        elif self.style == AuditLog.Style.CSV:
            await self._add_csv_entry(status)

    async def _add_plain_entry(self: AL, status: Status) -> None:
        """Append a plain text entry/status details to the audit log file."""
        entry = (
            f"{arrow.now().format('YYYY-MM-DD hh:mm:ss')} -"
            f" Removed {'poll' if status.get('poll', False) else 'status'} {status.get('url')}"
            f" created @ {status.get('created_at')}"
            f" with {status.get('visibility')} visibility, {len(status.get('media_attachments', []))} attachments."
            f" This status was reblogged {status.get('reblogs_count')} times and"
            f" favourited {status.get('favourites_count')} times."
            f" The status was {'pinned' if status.get('pinned') else 'not pinned'}.\n"
        )
        await self.audit_log.write(entry)

    async def _add_csv_entry(self: AL, status: Status) -> None:
        """Append a csv formatted status details to the audit log file."""
        csv_writer = aiocsv.writers.AsyncWriter(
            asyncfile=self.audit_log,
            quoting=csv.QUOTE_ALL,
        )
        record = [
            arrow.now(),
            "poll" if status.get("poll", False) else "status",
            status.get("url"),
            arrow.get(status.get("created_at", 0)).to(tz="local"),
            status.get("visibility"),
            len(status.get("media_attachments", [])),
            status.get("reblogs_count"),
            status.get("favourites_count"),
            status.get("pinned"),
        ]
        await csv_writer.writerow(row=record)

    @staticmethod
    async def _add_csv_header(csv_file: Any) -> None:
        """Write CSV header line to the audit log file.

        This method is normally called when an audit log file is being
        initially created before writing any other entries to the audit
        log; i.e. The record it writes will be the header record for the
        CSV file.
        """
        csv_writer = aiocsv.writers.AsyncWriter(
            asyncfile=csv_file, quoting=csv.QUOTE_ALL
        )
        record = [
            "date and time of deletion",
            "type of status,",
            "url of status before deletion",
            "date and time deleted status was created",
            "visibility",
            "# of media_attachments",
            "reblogs_count",
            "favourites_count",
            "pinned",
        ]
        await csv_writer.writerow(row=record)


def should_keep(  # noqa PLR0911
    status: Status, oldest_to_keep: arrow.Arrow, config: Configuration
) -> bool:
    """Determine if status should be kept."""
    status_created_at = arrow.get(cast(int, status.get("created_at")))
    if status_created_at >= oldest_to_keep:
        return True

    if config.bot.skip_deleting_bookmarked and bool(status.get("bookmarked")):
        return True

    if config.bot.skip_deleting_faved and bool(status.get("favourited")):
        return True

    if config.bot.skip_deleting_pinned and bool(status.get("pinned")):
        return True

    if config.bot.skip_deleting_poll and bool(status.get("poll")):
        return True

    if config.bot.skip_deleting_dm and status.get("visibility") == "direct":
        return True

    medias = status.get("media_attachments")
    if (
        config.bot.skip_deleting_media
        and isinstance(medias, list)
        and bool(len(medias))
    ):
        return True

    favourites = status.get("favourites_count", 0)
    if (
        config.bot.skip_deleting_faved_at_least
        and favourites >= config.bot.skip_deleting_faved_at_least
    ):
        return True

    reblogs = status.get("reblogs_count", 0)
    if (
        config.bot.skip_deleting_boost_at_least
        and reblogs >= config.bot.skip_deleting_boost_at_least
    ):
        return True

    if config.bot.skip_deleting_reactions_at_least:
        reactions_threshold = config.bot.skip_deleting_reactions_at_least
    else:
        reactions_threshold = 0

    return enough_reactions_to_keep(
        status=status,
        reactions_threshold=reactions_threshold,
    )


def enough_reactions_to_keep(status: Status, reactions_threshold: int) -> bool:
    """Check if emoji_reactions are reported in status and check of they are at least at the
    threshold for keeping the status.
    Note: Reactions are a Pleroma specific feature.

    :param status: Status being examined for reactions
    :param reactions_threshold: Minimum number of reactions to consider the check to be met

    :returns:
        True if the total number of reactions to status is equal to or more than the reactions_threshold.
        False in any other case, including if the status does not contain any reaction counts at all.
    """
    if not (pleroma := status.get("pleroma")):
        return False

    if not (reactions := pleroma.get("emoji_reactions")):
        return False

    logger.debug(
        "enough_reactions_to_keep - reactions_threshold: %s - emoji_reactions:\n%s",
        reactions_threshold,
        json.dumps(reactions, indent=4),
    )

    total_reactions = 0
    for reaction in reactions:
        if emo_count := reaction.get("count"):
            total_reactions += emo_count

    return total_reactions >= reactions_threshold



"""Fedinesia - deletes old statuses from fediverse accounts. This tool was previously
called MastodonAmnesia
Copyright (C) 2021, 2022, 2023  Mark S Burgunder.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import asyncio
import json
import logging
import sys
from math import ceil
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import aiohttp
import arrow
import typer
from aiohttp.client_exceptions import ClientError
from minimal_activitypub.client_2_server import ActivityPub
from minimal_activitypub.client_2_server import ActivityPubError
from minimal_activitypub.client_2_server import RatelimitError
from rich import print
from rich import traceback
from tqdm import tqdm
from tqdm import trange
from typing_extensions import Annotated

from . import Status
from . import __display_name__
from . import __version__
from .config import setup_shop
from .util import AuditLog
from .util import check_updates
from .util import should_keep

traceback.install(show_locals=True)
logger = logging.getLogger(__display_name__)
logger.setLevel(logging.DEBUG)


async def main(  # noqa: C901, PLR0912, PLR0913, PLR0915
    config_file: str,
    is_dry_run: bool,
    debug_log_file: Optional[str],
    audit_log_file: Optional[str],
    audit_log_style: Optional[str],
    batch_size: Optional[int],
    limit: Optional[int],
) -> None:
    """Perform app function."""
    config = await setup_shop(
        config_file=config_file,
        debug_log_file=debug_log_file,
    )

    oldest_to_keep = arrow.now().shift(seconds=-config.bot.delete_after)

    print(f"Welcome to {__display_name__} {__version__}")
    logger.debug("main -Welcome to %s %s", __display_name__, __version__)

    check_updates()

    try:
        session = aiohttp.ClientSession()

        instance = ActivityPub(
            instance=config.mastodon.instance,
            access_token=config.mastodon.access_token,
            session=session,
        )
        await instance.determine_instance_type()
        user_info = await instance.verify_credentials()
        print(
            f"We are removing statuses older than {oldest_to_keep} "
            f"from {config.mastodon.instance}@{user_info['username']}"
        )

        audit_log = None
        if audit_log_file:
            print(
                f"A record of all deleted statuses will be recorded in the audit log file at {audit_log_file}"
            )
            audit_log = await AuditLog.open(
                audit_log_file=audit_log_file,
                style=audit_log_style,
            )

        statuses = await instance.get_account_statuses(account_id=user_info["id"])
    except RatelimitError:
        print(
            f"RateLimited during startup, [red]Please wait until[/red] "
            f"{instance.ratelimit_reset} before trying again"
        )
        sys.exit(429)
    except (ClientError, ActivityPubError):
        logger.exception("!!! Cannot continue.")
        sys.exit(100)

    statuses_to_delete: List[Status] = []
    title = "Finding statuses to delete"
    progress_bar = tqdm(
        desc=f"{title:.<60}",
        ncols=120,
        unit="statuses",
        position=0,
        bar_format="{l_bar} {n_fmt} at {rate_fmt}",
    )
    while True:
        try:
            for status in statuses:
                logger.debug(
                    "Processing status: %s from %s",
                    status.get("url"),
                    arrow.get(status.get("created_at")).to(tz="local"),
                )
                logger.debug(
                    "Oldest to keep vs status created at %s > %s",
                    oldest_to_keep,
                    arrow.get(status.get("created_at")).to(tz="local"),
                )

                if should_keep(
                    status=status,
                    oldest_to_keep=oldest_to_keep,
                    config=config,
                ):
                    logger.info(
                        "Not deleting status: "
                        "Bookmarked: %s - "
                        "My Fav: %s - "
                        "Pinned: %s - "
                        "Poll: %s - "
                        "Attachements: %s - "
                        "Faved: %s - "
                        "Boosted: %s - "
                        "DM: %s -+- "
                        "Created At: %s -+- "
                        "%s",
                        status.get("bookmarked"),
                        status.get("favourited"),
                        status.get("pinned"),
                        (status.get("poll") is not None),
                        len(status.get("media_attachments")),
                        status.get("favourites_count"),
                        status.get("reblogs_count"),
                        (status.get("visibility") == "direct"),
                        arrow.get(status.get("created_at")).to(tz="local"),
                        status.get("url"),
                    )

                elif limit and len(statuses_to_delete) >= limit:
                    break

                else:
                    statuses_to_delete.append(status)

                progress_bar.update()

            if limit and len(statuses_to_delete) >= limit:
                break

            # Get More statuses if available:
            logger.debug("Main - get next batch of statuses if available.")
            logger.debug("Main - instance.pagination: %s", instance.pagination)
            if (
                instance.pagination["next"]["max_id"]
                or instance.pagination["next"]["min_id"]
            ):
                statuses = await instance.get_account_statuses(
                    account_id=user_info["id"],
                    max_id=instance.pagination["next"]["max_id"],
                    min_id=instance.pagination["next"]["min_id"],
                )
                logger.debug("Main - scrolling - len(statuses): %s", len(statuses))
                if len(statuses) == 0:
                    break
            else:
                break

        except RatelimitError:
            await sleep_off_ratelimiting(instance=instance)

    progress_bar.close()

    total_statuses_to_delete = len(statuses_to_delete)
    logger.debug(
        "Main - start deleting - total_statuses_to_delete: %s", total_statuses_to_delete
    )
    for status in statuses_to_delete:
        logger.debug(
            "Start of deleting - status to delete: %s @ %s",
            status["id"],
            status["url"],
        )

    # If dry-run has been specified, print out list of statuses that would be deleted
    if is_dry_run:
        print("\n--dry-run or -d specified. [yellow][bold]No statuses will be deleted")
        for status in statuses_to_delete:
            print(
                f"[red]Would[/red] delete status"
                f" {status.get('url')} from {status.get('created_at')}"
            )
        print(f"Total of {total_statuses_to_delete} statuses would be deleted.")

    # Dry-run has not been specified... delete statuses!
    else:
        await delete_statuses(
            instance=instance,
            statuses_to_delete=statuses_to_delete,
            audit=audit_log,
            batch_size=batch_size,
        )
        print(
            f"All old statuses deleted! Total of {total_statuses_to_delete} statuses deleted"
        )

    if audit_log:
        await audit_log.close()
    await session.close()


async def delete_statuses(
    instance: ActivityPub,
    statuses_to_delete: List[Status],
    audit: Optional[AuditLog],
    batch_size: Optional[int],
) -> None:
    """Delete all statuses that should be deleted."""
    title = "Deleting statuses"
    total_statuses_to_delete = len(statuses_to_delete)
    if total_statuses_to_delete > 0:
        with tqdm(
            desc=f"{title:.<60}",
            ncols=120,
            total=total_statuses_to_delete,
            unit="statuses",
            position=0,
            bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} at {rate_fmt}",
        ) as progress_bar:
            while len(statuses_to_delete) > 0:
                tasks = []
                for status in statuses_to_delete:
                    tasks.append(
                        delete_single_status(
                            status=status, instance=instance, audit=audit
                        )
                    )
                    if batch_size and len(tasks) >= batch_size:
                        break

                responses = await asyncio.gather(*tasks)

                # Filter out any responses that are None...
                # Those encountered Rate Limiting
                deleted_statuses = [
                    status for status in responses if status is not None
                ]
                logger.debug(
                    "delete_statuses - len(deleted_statuses): %s",
                    len(deleted_statuses),
                )

                progress_bar.update(len(deleted_statuses))

                if len(deleted_statuses) < len(statuses_to_delete):
                    await sleep_off_ratelimiting(instance=instance)

                for status in deleted_statuses:
                    statuses_to_delete.remove(status)


async def sleep_off_ratelimiting(
    instance: ActivityPub,
) -> None:
    """Wait for rate limiting to be over."""
    logger.debug(
        "sleep_off_ratelimiting - Rate limited: Limit: %s - resetting at: %s",
        instance.ratelimit_remaining,
        instance.ratelimit_reset,
    )
    reset_at = arrow.get(instance.ratelimit_reset).datetime
    now = arrow.now().datetime
    need_to_wait = ceil((reset_at - now).total_seconds())

    logger.info(
        "Need to wait %s seconds (until %s) to let server 'cool down'",
        need_to_wait,
        arrow.get(instance.ratelimit_reset),
    )
    bar_title = "Waiting to let server 'cool-down'"
    for _i in trange(
        need_to_wait,
        desc=f"{bar_title:.<60}",
        unit="s",
        ncols=120,
        bar_format="{desc}: {percentage:3.0f}%|{bar}| Eta: {remaining} - Elapsed: {elapsed}",
        position=1,
    ):
        await asyncio.sleep(1)


async def delete_single_status(
    status: Status,
    instance: ActivityPub,
    audit: Optional[AuditLog],
) -> Optional[Dict[str, Any]]:
    """Delete  single status."""
    logger.debug(
        "delete_single_status(status=%s, instance=%s)",
        status["id"],
        instance.instance,
    )
    return_status: Optional[Status] = status
    try:
        await instance.delete_status(status=status)
        logger.info(
            "delete_single_status - Deleted status %s from %s",
            status.get("url"),
            status.get("created_at"),
        )
        if audit:
            await audit.add_entry(status=status)
    except RatelimitError:
        logger.debug(
            "delete_single_status - status id = %s - ratelimit_remaining = %s - ratelimit_reset = %s",
            status["id"],
            instance.ratelimit_remaining,
            instance.ratelimit_reset,
        )
        return_status = None
    except ActivityPubError as error:
        logger.debug(
            "delete_single_status - encountered error: %s",
            error,
        )
        logger.debug("delete_single_status - status: %s", json.dumps(status, indent=4))
        raise error

    return return_status


def start() -> None:
    """Start app."""
    typer.run(typer_async_shim)


def typer_async_shim(  # noqa PLR0913
    config_file: Annotated[str, typer.Option("-c", "--config-file")] = "config.json",
    debug_log_file: Optional[str] = None,
    audit_log_file: Annotated[
        Optional[str], typer.Option("-a", "--audit-log-file")
    ] = None,
    audit_log_style: Optional[AuditLog.Style] = AuditLog.Style.PLAIN,
    batch_size: Annotated[Optional[int], typer.Option("-b", "--batch-size")] = None,
    limit: Annotated[Optional[int], typer.Option("-l", "--limit")] = None,
    dry_run: Annotated[bool, typer.Option("-d", "--dry-run")] = False,
) -> None:
    """Delete fediverse history. For more information look at https://codeberg.org/MarvinsMastodonTools/fedinesia."""
    asyncio.run(
        main(
            config_file=config_file,
            is_dry_run=dry_run,
            debug_log_file=debug_log_file,
            audit_log_file=audit_log_file,
            audit_log_style=audit_log_style,
            batch_size=batch_size,
            limit=limit,
        )
    )



"""Package 'fedinesia' level definitions."""
import sys
from typing import Any
from typing import Dict
from typing import Final

__version__: Final[str] = "2.5.4"
__display_name__: Final[str] = "Fedinesia"
__package_name__: Final[str] = __display_name__.lower()
USER_AGENT: Final[
    str
] = f"{__display_name__}_v{__version__}_Python_{sys.version.split()[0]}"

CLIENT_WEBSITE: Final[str] = "https://codeberg.org/MarvinsMastodonTools/fedinesia"

Status = Dict[str, Any]
